from datetime import datetime
from functools import partial
import pandas as pd
from opa_tools.config import get_benchmark_dir
from opa_tools.utils import process_dct
import yaml
import warnings


BENCHMARK_DIR = get_benchmark_dir()
_process_dct = partial(process_dct, wrapper_fun=lambda x: [x])


def _select_directories_by_dates(dates):
    directories = sorted(path.name for path in BENCHMARK_DIR.iterdir())  # type: ignore
    if dates:
        directories = list(
            filter(
                lambda directory: any(directory.startswith(date) for date in dates),
                directories,
            )
        )
    return directories


def _get_exclude_fun(exclude, directories):
    """Takes in arguments exclude and directories, and
    returns a function that takes in a path and returns a bool.
    """
    if exclude is None:
        return lambda _: False
    else:
        if len(directories) == 1:
            if isinstance(exclude, str):
                return lambda path: exclude == path.name
            else:
                raise NotImplementedError(
                    "Still must use exclude with a str and together with single date."
                )
        else:
            raise NotImplementedError(
                "Still must use exclude with a str and together with single date."
            )


def _get_interval_fun(interval):
    def is_in_interval(benchmark_dir):
        try:
            with open(benchmark_dir / "date.txt", "r") as datefile:
                date = datetime.fromisoformat(datefile.read())
            return interval[0] <= date < interval[1]
        except FileNotFoundError:
            warnings.warn(f"File {benchmark_dir / 'date.txt'} not found.A")
            return False

    if interval is None:
        return lambda x: True
    return is_in_interval


def get_benchmark_iterator(dates=None, exclude=None, interval=None):
    directories = _select_directories_by_dates(dates)
    exclude_fun = _get_exclude_fun(exclude, directories)
    interval_fun = _get_interval_fun(interval)
    for date in directories:
        directory_date = BENCHMARK_DIR / date
        for benchmark in sorted(directory_date.iterdir(), key=lambda x: x.name):
            if exclude_fun(directory_date / benchmark):
                continue
            if not interval_fun(directory_date / benchmark):
                continue
            yield benchmark


def get_environ_df(dates=None, exclude=None, interval=None):
    environ_df = pd.DataFrame()
    for benchmark_dir in get_benchmark_iterator(dates, exclude, interval):
        benchmark_name = benchmark_dir.name
        with (
            open(benchmark_dir / "date.txt", "r") as date_file,
            open(benchmark_dir / "environ.yaml", "r") as environ_file,
        ):
            try:
                environ_dict = yaml.safe_load(environ_file)
            except yaml.constructor.ConstructorError:
                continue
            dct = {
                "date": date_file.read(),
                "bname": benchmark_name,
                **_process_dct(environ_dict),
            }
            entry = pd.DataFrame.from_dict(dct)
            environ_df = pd.concat([environ_df, entry], ignore_index=True)
    return environ_df


def get_request_df(dates=None, exclude=None, interval=None):
    request_df = pd.DataFrame()
    for benchmark_dir in get_benchmark_iterator(dates, exclude, interval):
        benchmark_name = benchmark_dir.name
        with (
            open(benchmark_dir / "date.txt", "r") as date_file,
            open(
                benchmark_dir / "additional" / "gsvrequest.yaml", "r"
            ) as gsv_request_file,
            open(
                benchmark_dir / "additional" / "oparequest.yaml", "r"
            ) as opa_request_file,
        ):
            gsv_dict = yaml.safe_load(gsv_request_file)
            opa_dict = yaml.safe_load(opa_request_file)
            dct = {
                "date": date_file.read(),
                "bname": benchmark_name,
                **_process_dct(gsv_dict, prefix="gsv_"),
                **_process_dct(opa_dict, prefix="opa_"),
            }
            entry = pd.DataFrame.from_dict(dct)
            request_df = pd.concat([request_df, entry], ignore_index=True)
    return request_df
