from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union
import yaml

from opa_tools.config import get_request_dir
from .transformation import flatten
from .utils import RequestIdentifierError


REQUEST_DIR = get_request_dir()
ALLOWED_KINDS = (
    "gsv",
    "opa",
)

def _normalize_request_name(name: Any) -> Tuple[str, Union[str, None]]:
    if isinstance(name, float):
        raise NotImplementedError(
            "Passing request name as float not implemented yet. "
            "Please, pass as a string with a dot."
        )
    if isinstance(name, int):
        return str(name), None
    if isinstance(name, str):
        if "." in name:
            name, subrequest = name.split(".", 1)
        else:
            subrequest = None
        return name, subrequest
    raise TypeError(
        "Request (and subrequest) names must be expressed as "
        f"int, float or str, not {type(name)}."
    )

def _validate_month(month: str):
    if not 1 <= int(month) <= 12:
        raise ValueError(f"Bad value for month: {month}.")

def _normalize_request_month(month: str | int | None, identifier: Any) -> str:
    now = datetime.now()
    if month is None:
        return now.strftime("%y%m")
    month = str(month)
    if month == "*":
        return month
    if not month.isdigit():
        raise ValueError(f"Value for month must be a number, but was '{month}'.")
    if len(month) == 2:
        _validate_month(month)
        return now.strftime("%y") + month
    elif len(month) == 4:
        _validate_month(month[2:4])
        return month
    raise RequestIdentifierError(identifier)


def _normalize_request_identifier(
    identifier: Any,
    subrequest: Optional[Any] = None,
    month: Optional[Union[str, int]] = None,
    kind: str = "*"
) -> Tuple[str, Union[str, None], str, str]:
    # Split identifier
    if isinstance(identifier, str) and "/" in identifier:
        month, name = identifier.split("/", 1)
    else:
        name = identifier
        month = _normalize_request_month(month, identifier)
    # Normalize triplet
    name, subrequest = _normalize_request_name(name)
    month = _normalize_request_month(month, identifier)
    if kind not in ["*", *ALLOWED_KINDS]:
        raise ValueError(f"Wrong kind: '{kind}', must be in {ALLOWED_KINDS}.")
    return name, subrequest, month, kind


def _find_zero_padded_subdir(path: Path, subdir: str, max_digits=12):
    if len(subdir) >= max_digits:
        raise RuntimeError(
            f"Subdirectory {subdir} too long to be found in {path}"
        )
    for n_zeros in range(0, max_digits - len(subdir)):
        cur_path = path / subdir.zfill(n_zeros)
        if cur_path.exists():
            return cur_path
    raise FileNotFoundError(
        f"Directory {path / subdir} (or any other zero-padded variant, "
        f"up to {max_digits} digits) not found for request."
    )


def _get_single_request(name: str, subrequest: Union[str, None], month: str, kind: str) -> Dict:
    if kind not in ("opa", "gsv"):
        raise AssertionError(f"kind must be either 'opa' or 'gsv', but was '{kind}'.")
    request_dir = _find_zero_padded_subdir(REQUEST_DIR / month, name)
    if subrequest:
        request_dir = _find_zero_padded_subdir(request_dir, subrequest)
    request_path = request_dir / f"{kind}.yml"
    if request_path.exists():
        with open(request_path, "r", encoding="utf-8") as rf:
            return yaml.safe_load(rf)
    if any(subpath.is_dir() for subpath in request_dir.iterdir()):
        raise NotImplementedError(
            "Ability to fetch lists of subrequests not implemented yet. "
            "Please, pass subrequests manually (syntax: YYMM/REQID.SUBREQID)"
        )
    raise FileNotFoundError(
        f"Directory {request_dir} exists but does not contain requests or subdirectories."
    )


def get_request_identifiers(month: str = "*"):
    if month == "*":
        output = []
        for month in sorted(path.name for path in REQUEST_DIR.iterdir()):
            output.extend(get_request_identifiers(month=month))
        return output
    output = []
    for path in sorted((REQUEST_DIR / month).iterdir()):
        output.append(path.parent.name + "/" + path.name)
    return output


def _get_requests(
    name: str,
    subrequest: Union[str, None],
    month: str,
    kind: str,
    flat: bool,

) -> List[Dict]:
    if kind == "*":
        output = [
            _get_single_request(name=name, subrequest=subrequest, month=month, kind="gsv"),
            _get_single_request(name=name, subrequest=subrequest, month=month, kind="opa"),
        ]
        if flat:
            return flatten(*output)
        return output
    return [_get_single_request(name=name, subrequest=subrequest, month=month, kind=kind)]

def get_request(
    identifier: Optional[Union[str, int, float]] = None,
    /,
    subrequest: Optional[str] = None,
    month: Optional[str] = None,
    kind: str = "*",
    *,
    flat=False,
    single_as_list=False,
) -> Union[List[Dict], Dict]:
    if identifier is None:
        if month is None:
            month = "*"
        output = []
        for identifier in get_request_identifiers(month=month):
            output.append(
                get_request(identifier, kind=kind, flat=flat, single_as_list=True)
            )
        return output
    name, subrequest, month, kind = _normalize_request_identifier(identifier, subrequest, month, kind)
    requests = _get_requests(name=name, subrequest=subrequest, month=month, kind=kind, flat=flat)
    if len(requests) == 1 and single_as_list:
        return requests[0]
    return requests

